"""Class definition for bijection."""
from collections.abc import Mapping, MutableMapping

__all__ = ('bijection', )


class bijection(MutableMapping):
	"""A one-to-one onto mapping, a dict with unique values.

	.. automethod:: __init__
	"""

	def __init__(self, iterable=None, **kwarg):
		"""Create a bijection from an iterable.

		Matches dict.__init__.
		"""
		self._data = {}
		self.__inverse = self.__new__(bijection)
		self.__inverse._data = {}
		self.__inverse.__inverse = self
		if iterable is not None:
			if isinstance(iterable, Mapping):
				for key, value in iterable.items():
					self[key] = value
			else:
				for pair in iterable:
					key, value = pair
					self[key] = value
		for key, value in kwarg.items():
			self[key] = value

	def __repr__(self):
		if len(self._data) == 0:
			return '{0}()'.format(self.__class__.__name__)
		else:
			repr_format = '{class_name}({values!r})'
			return repr_format.format(
				class_name=self.__class__.__name__,
				values=self._data,
				)

	@property
	def inverse(self):
		"""Return the inverse of this bijection."""
		return self.__inverse

	# Required for MutableMapping
	def __len__(self):
		return len(self._data)

	# Required for MutableMapping
	def __getitem__(self, key):
		return self._data[key]

	# Required for MutableMapping
	def __setitem__(self, key, value):
		if key in self:
			del self.inverse._data[self[key]]
		if value in self.inverse:
			del self._data[self.inverse[value]]
		self._data[key] = value
		self.inverse._data[value] = key

	# Required for MutableMapping
	def __delitem__(self, key):
		value = self._data.pop(key)
		del self.inverse._data[value]

	# Required for MutableMapping
	def __iter__(self):
		return iter(self._data)

	def __contains__(self, key):
		return key in self._data

	def clear(self):
		"""Remove everything from this bijection."""
		self._data.clear()
		self.inverse._data.clear()

	def copy(self):
		"""Return a copy of this bijection."""
		return bijection(self)

	def items(self):
		"""See Mapping.items."""
		return self._data.items()

	def keys(self):
		"""See Mapping.keys."""
		return self._data.keys()

	def values(self):
		"""See Mapping.values."""
		return self.inverse.keys()

	def __eq__(self, other):
		return isinstance(other, bijection) and self._data == other._data



"""Setlist class definitions."""
import random as random_
from collections.abc import (
	Hashable,
	MutableSequence,
	MutableSet,
	Sequence,
	Set,
	)

from . import _util

__all__ = ('SetList', 'setlist', 'frozensetlist')


class SetList(Sequence, Set):
	"""A setlist is an ordered `Collection` of unique elements.

	`SetList` is the superclass of `setlist` and `frozensetlist`. It is
	immutable and unhashable.
	"""

	def __init__(self, iterable=None, raise_on_duplicate=False):
		"""Create a setlist, initializing from iterable if present.

		Args:
			iterable (Iterable): Values to initialize the setlist with.
			raise_on_duplicate: Raise a ValueError if any duplicate values
				are present.
		"""
		self._list = list()
		self._dict = dict()
		if iterable:
			if raise_on_duplicate:
				self._extend(iterable)
			else:
				self._update(iterable)

	def __repr__(self):
		if len(self) == 0:
			return '{0}()'.format(self.__class__.__name__)
		else:
			repr_format = '{class_name}({values!r})'
			return repr_format.format(
				class_name=self.__class__.__name__,
				values=tuple(self),
				)

	# Convenience methods

	def _fix_neg_index(self, index):
		if index < 0:
			index += len(self)
		if index < 0:
			raise IndexError('index is out of range')
		return index

	def _fix_end_index(self, index):
		if index is None:
			return len(self)
		else:
			return self._fix_neg_index(index)

	def _append(self, value):
		# Checking value in self will check that value is Hashable
		if value in self:
			raise ValueError('Value "%s" already present' % str(value))
		else:
			self._dict[value] = len(self)
			self._list.append(value)

	def _extend(self, values):
		new_values = set()
		for value in values:
			if value in new_values:
				raise ValueError('New values contain duplicates')
			elif value in self:
				raise ValueError('New values contain elements already present in self')
			else:
				new_values.add(value)
		for value in values:
			self._dict[value] = len(self)
			self._list.append(value)

	def _add(self, item):
		if item not in self:
			self._dict[item] = len(self)
			self._list.append(item)

	def _update(self, values):
		for value in values:
			if value not in self:
				self._dict[value] = len(self)
				self._list.append(value)

	@classmethod
	def _from_iterable(cls, it, **kwargs):
		return cls(it, **kwargs)

	# Implement Container
	def __contains__(self, value):
		return value in self._dict

	# Iterable we get by inheriting from Sequence

	# Implement Sized
	def __len__(self):
		return len(self._list)

	# Implement Sequence
	def __getitem__(self, index):
		if isinstance(index, slice):
			return self._from_iterable(self._list[index])
		return self._list[index]

	def count(self, value):
		"""Return the number of occurrences of value in self.

		This runs in O(1)

		Args:
			value: The value to count
		Returns:
			int: 1 if the value is in the setlist, otherwise 0
		"""
		if value in self:
			return 1
		else:
			return 0

	def index(self, value, start=0, end=None):
		"""Return the index of value between start and end.

		By default, the entire setlist is searched.

		This runs in O(1)

		Args:
			value: The value to find the index of
			start (int): The index to start searching at (defaults to 0)
			end (int): The index to stop searching at (defaults to the end of the list)
		Returns:
			int: The index of the value
		Raises:
			ValueError: If the value is not in the list or outside of start - end
			IndexError: If start or end are out of range
		"""
		try:
			index = self._dict[value]
		except KeyError:
			raise ValueError
		else:
			start = self._fix_neg_index(start)
			end = self._fix_end_index(end)
			if start <= index < end:
				return index
			else:
				raise ValueError

	@classmethod
	def _check_type(cls, other, operand_name):
		if not isinstance(other, SetList):
			message = (
				"unsupported operand type(s) for {operand_name}: "
				"'{self_type}' and '{other_type}'").format(
					operand_name=operand_name,
					self_type=cls,
					other_type=type(other),
					)
			raise TypeError(message)

	def __add__(self, other):
		self._check_type(other, '+')
		out = self.copy()
		out._extend(other)
		return out

	# Implement Set

	def issubset(self, other):
		"""Report whether another set contains this set."""
		return self <= other

	def issuperset(self, other):
		"""Report whether this set contains another set."""
		return self >= other

	def union(self, other):
		"""Return the union of sets as a new set.

		(i.e. all elements that are in either set.)
		"""
		out = self.copy()
		out._update(other)
		return out

	def intersection(self, other):
		"""Return the intersection of two sets as a new set.

		(i.e. all elements that are in both sets.)
		"""
		other = set(other)
		return self._from_iterable(item for item in self if item in other)

	def difference(self, other):
		"""Return the difference of two or more sets as a new set.

		(i.e. all elements that are in this set but not the others.)
		"""
		other = set(other)
		return self._from_iterable(item for item in self if item not in other)

	def symmetric_difference(self, other):
		"""Return the symmetric difference (disjuntive union) of two sets.

		(i.e. all elements that are in one set but not both.)
		"""
		return self.union(other) - self.intersection(other)

	def __sub__(self, other):
		self._check_type(other, '-')
		return self.difference(other)

	def __and__(self, other):
		self._check_type(other, '&')
		return self.intersection(other)

	def __or__(self, other):
		self._check_type(other, '|')
		return self.union(other)

	def __xor__(self, other):
		self._check_type(other, '^')
		return self.symmetric_difference(other)

	# Comparison

	def __eq__(self, other):
		if not isinstance(other, SetList):
			return False
		if not len(self) == len(other):
			return False
		for self_elem, other_elem in zip(self, other):
			if self_elem != other_elem:
				return False
		return True

	def __ne__(self, other):
		return not (self == other)

	# New methods

	def sub_index(self, sub, start=0, end=None):
		"""Return the index of a subsequence.

		This runs in O(len(sub))

		Args:
			sub (Sequence): An Iterable to search for
			start (int): The index at which to start the search
			end (int): The index at which to end the search
		Returns:
			int: The index of the first element of sub
		Raises:
			ValueError: If sub isn't a subsequence
			TypeError: If sub isn't iterable
			IndexError: If start or end are out of range
		"""
		start_index = self.index(sub[0], start, end)
		end = self._fix_end_index(end)
		if start_index + len(sub) > end:
			raise ValueError
		for i in range(1, len(sub)):
			if sub[i] != self[start_index + i]:
				raise ValueError
		return start_index

	def copy(self):
		"""Return a shallow copy of the setlist."""
		return self.__class__(self)


class setlist(SetList, MutableSequence, MutableSet):
	"""A mutable (unhashable) setlist.

	.. automethod:: __init__
	"""

	def __str__(self):
		return '{[%s}]' % ', '.join(repr(v) for v in self)

	# Helper methods
	def _delete_all(self, elems_to_delete, raise_errors):
		indices_to_delete = set()
		for elem in elems_to_delete:
			try:
				elem_index = self._dict[elem]
			except KeyError:
				if raise_errors:
					raise ValueError('Passed values contain elements not in self')
			else:
				if elem_index in indices_to_delete:
					if raise_errors:
						raise ValueError('Passed vales contain duplicates')
				indices_to_delete.add(elem_index)
		self._delete_values_by_index(indices_to_delete)

	def _delete_values_by_index(self, indices_to_delete):
		deleted_count = 0
		for i, elem in enumerate(self._list):
			if i in indices_to_delete:
				deleted_count += 1
				del self._dict[elem]
			else:
				new_index = i - deleted_count
				self._list[new_index] = elem
				self._dict[elem] = new_index
		# Now remove deleted_count items from the end of the list
		if deleted_count:
			self._list = self._list[:-deleted_count]

	# Set/Sequence agnostic
	def pop(self, index=-1):
		"""Remove and return the item at index."""
		value = self._list.pop(index)
		del self._dict[value]
		return value

	def clear(self):
		"""Remove all elements from self."""
		self._dict = dict()
		self._list = list()

	# Implement MutableSequence
	def __setitem__(self, index, value):
		if isinstance(index, slice):
			old_values = self[index]
			for v in value:
				if v in self and v not in old_values:
					raise ValueError
			self._list[index] = value
			self._dict = {}
			for i, v in enumerate(self._list):
				self._dict[v] = i
		else:
			index = self._fix_neg_index(index)
			old_value = self._list[index]
			if value in self:
				if value == old_value:
					return
				else:
					raise ValueError
			del self._dict[old_value]
			self._list[index] = value
			self._dict[value] = index

	def __delitem__(self, index):
		if isinstance(index, slice):
			indices_to_delete = set(self.index(e) for e in self._list[index])
			self._delete_values_by_index(indices_to_delete)
		else:
			index = self._fix_neg_index(index)
			value = self._list[index]
			del self._dict[value]
			for elem in self._list[index + 1:]:
				self._dict[elem] -= 1
			del self._list[index]

	def insert(self, index, value):
		"""Insert value at index.

		Args:
			index (int): Index to insert value at
			value: Value to insert
		Raises:
			ValueError: If value already in self
			IndexError: If start or end are out of range
		"""
		if value in self:
			raise ValueError
		index = self._fix_neg_index(index)
		self._dict[value] = index
		for elem in self._list[index:]:
			self._dict[elem] += 1
		self._list.insert(index, value)

	def append(self, value):
		"""Append value to the end.

		Args:
			value: Value to append
		Raises:
			ValueError: If value alread in self
			TypeError: If value isn't hashable
		"""
		self._append(value)

	def extend(self, values):
		"""Append all values to the end.

		If any of the values are present, ValueError will
		be raised and none of the values will be appended.

		Args:
			values (Iterable): Values to append
		Raises:
			ValueError: If any values are already present or there are duplicates
				in the passed values.
			TypeError: If any of the values aren't hashable.
		"""
		self._extend(values)

	def __iadd__(self, values):
		"""Add all values to the end of self.

		Args:
			values (Iterable): Values to append
		Raises:
			ValueError: If any values are already present
		"""
		self._check_type(values, '+=')
		self.extend(values)
		return self

	def remove(self, value):
		"""Remove value from self.

		Args:
			value: Element to remove from self
		Raises:
			ValueError: if element is already present
		"""
		try:
			index = self._dict[value]
		except KeyError:
			raise ValueError('Value "%s" is not present.')
		else:
			del self[index]

	def remove_all(self, elems_to_delete):
		"""Remove all elements from elems_to_delete, raises ValueErrors.

		See Also:
			discard_all
		Args:
			elems_to_delete (Iterable): Elements to remove.
		Raises:
			ValueError: If the count of any element is greater in
				elems_to_delete than self.
			TypeError: If any of the values aren't hashable.
		"""
		self._delete_all(elems_to_delete, raise_errors=True)

	def reverse(self):
		"""Reverse the setlist in-place."""
		self._list.reverse()
		for index, item in enumerate(self._list):
			self._dict[item] = index

	# Implement MutableSet

	def add(self, item):
		"""Add an item.

		Note:
			This does not raise a ValueError for an already present value like
			append does. This is to match the behavior of set.add
		Args:
			item: Item to add
		Raises:
			TypeError: If item isn't hashable.
		"""
		self._add(item)

	def update(self, values):
		"""Add all values to the end.

		If any of the values are present, silently ignore
		them (as opposed to extend which raises an Error).

		See also:
			extend
		Args:
			values (Iterable): Values to add
		Raises:
			TypeError: If any of the values are unhashable.
		"""
		self._update(values)

	def discard_all(self, elems_to_delete):
		"""Discard all the elements from elems_to_delete.

		This is much faster than removing them one by one.
		This runs in O(len(self) + len(elems_to_delete))

		Args:
			elems_to_delete (Iterable): Elements to discard.
		Raises:
			TypeError: If any of the values aren't hashable.
		"""
		self._delete_all(elems_to_delete, raise_errors=False)

	def discard(self, value):
		"""Discard an item.

		Note:
			This does not raise a ValueError for a missing value like remove does.
			This matches the behavior of set.discard
		"""
		try:
			self.remove(value)
		except ValueError:
			pass

	def difference_update(self, other):
		"""Update self to include only the difference with other."""
		other = set(other)
		indices_to_delete = set()
		for i, elem in enumerate(self):
			if elem in other:
				indices_to_delete.add(i)
		if indices_to_delete:
			self._delete_values_by_index(indices_to_delete)

	def intersection_update(self, other):
		"""Update self to include only the intersection with other."""
		other = set(other)
		indices_to_delete = set()
		for i, elem in enumerate(self):
			if elem not in other:
				indices_to_delete.add(i)
		if indices_to_delete:
			self._delete_values_by_index(indices_to_delete)

	def symmetric_difference_update(self, other):
		"""Update self to include only the symmetric difference with other."""
		other = setlist(other)
		indices_to_delete = set()
		for i, item in enumerate(self):
			if item in other:
				indices_to_delete.add(i)
		for item in other:
			self.add(item)
		self._delete_values_by_index(indices_to_delete)

	def __isub__(self, other):
		self._check_type(other, '-=')
		self.difference_update(other)
		return self

	def __iand__(self, other):
		self._check_type(other, '&=')
		self.intersection_update(other)
		return self

	def __ior__(self, other):
		self._check_type(other, '|=')
		self.update(other)
		return self

	def __ixor__(self, other):
		self._check_type(other, '^=')
		self.symmetric_difference_update(other)
		return self

	# New methods
	def shuffle(self, random=None):
		"""Shuffle all of the elements in self in place.

		Args:
			random: A function returning a random float in [0.0, 1.0). If none
				is passed, the default from `random.shuffle` will be used.
		"""
		random_.shuffle(self._list, random=random)
		for i, elem in enumerate(self._list):
			self._dict[elem] = i

	def sort(self, *args, **kwargs):
		"""Sort this setlist in place."""
		self._list.sort(*args, **kwargs)
		for index, value in enumerate(self._list):
			self._dict[value] = index

	def swap(self, i, j):
		"""Swap the values at indices i & j.

		.. versionadded:: 1.1
		"""
		i = self._fix_neg_index(i)
		j = self._fix_neg_index(j)
		self._list[i], self._list[j] = self._list[j], self._list[i]
		self._dict[self._list[i]] = i
		self._dict[self._list[j]] = j


class frozensetlist(SetList, Hashable):
	"""An immutable (hashable) setlist.

	.. automethod:: __init__
	"""

	def __hash__(self):
		if not hasattr(self, '_hash_value'):
			self._hash_value = _util.hash_iterable(self)
		return self._hash_value



"""collections_extended contains a few extra basic data structures."""
from collections.abc import Collection

from .bags import bag, frozenbag, UniqueElementsView, CountsView, Bag  # noqa
from .bijection import bijection
from .indexed_dict import IndexedDict
from .range_map import RangeMap, MappedRange  # noqa
from .setlists import frozensetlist, setlist, SetList

__all__ = (
	'Collection',
	'Bag',
	'bag',
	'frozenbag',
	'bijection',
	'IndexedDict',
	'RangeMap',
	'SetList',
	'frozensetlist',
	'setlist',
	)


def collection(iterable=None, mutable=True, ordered=False, unique=False):
	"""Return a :class:`Collection` with the specified properties.

	Args:
		iterable (Iterable): collection to instantiate new collection from.
		mutable (bool): Whether or not the new collection is mutable.
		ordered (bool): Whether or not the new collection is ordered.
		unique (bool): Whether or not the new collection contains only unique values.
	"""
	if iterable is None:
		iterable = tuple()
	if unique:
		if ordered:
			if mutable:
				return setlist(iterable)
			else:
				return frozensetlist(iterable)
		else:
			if mutable:
				return set(iterable)
			else:
				return frozenset(iterable)
	else:
		if ordered:
			if mutable:
				return list(iterable)
			else:
				return tuple(iterable)
		else:
			if mutable:
				return bag(iterable)
			else:
				return frozenbag(iterable)



"""util functions for collections_extended."""
from functools import wraps
import textwrap
import warnings

__all__ = ('hash_iterable', 'deprecated', 'deprecation_warning')


def hash_iterable(it):
	"""Perform a O(1) memory hash of an iterable of arbitrary length.

	hash(tuple(it)) creates a temporary tuple containing all values from it
	which could be a problem if it is large.

	See discussion at:
	https://groups.google.com/forum/#!msg/python-ideas/XcuC01a8SYs/e-doB9TbDwAJ
	"""
	hash_value = hash(type(it))
	for value in it:
		hash_value = hash((hash_value, value))
	return hash_value


def deprecation_warning(msg):
	"""Raise a deprecation warning."""
	warnings.warn(msg, category=DeprecationWarning, stacklevel=2)


def deprecated(msg, dep_version):
	"""Decorate a function, method or class to mark as deprecated.

	Raise DeprecationWarning and add a deprecation notice to the docstring.

	Args:
		msg: The message to document
		dep_version: The version in which this was deprecated
	See:
		https://www.sphinx-doc.org/en/master/usage/restructuredtext/directives.html#directive-deprecated

	"""
	def wrapper(func):
		docstring = func.__doc__ or ''
		docstring_msg = '.. deprecated:: {version} {msg}'.format(
			version=dep_version,
			msg=msg,
			)
		if docstring:
			# We don't know how far to indent this message
			# so instead we just dedent everything.
			string_list = docstring.splitlines()
			first_line = string_list[0]
			remaining = textwrap.dedent(''.join(string_list[1:]))
			docstring = '\n'.join([
				first_line,
				remaining,
				'',
				docstring_msg,
				])
		else:
			docstring = docstring_msg
		func.__doc__ = docstring

		@wraps(func)
		def inner(*args, **kwargs):
			deprecation_warning(msg)
			return func(*args, **kwargs)

		return inner

	return wrapper



"""RangeMap class definition."""
from abc import ABCMeta, abstractmethod
from bisect import bisect_left, bisect_right
from collections.abc import Collection, Mapping, Set

from .sentinel import NOT_SET


class MappedRange:
	"""Represents a subrange of a RangeMap.

	This is a glorified namedtuple.

	.. automethod:: __init__
	"""

	__slots__ = ('start', 'stop', 'value')

	def __init__(self, start, stop, value):
		"""Create a mapped range.

		Args:
			start: The start of the range, inclusive.
			stop: The end of the range, exclusive.
			value: The mapped value.
		"""
		self.start = start
		self.stop = stop
		self.value = value

	# Implement __iter__ so we can unpack this
	def __iter__(self):
		yield self.start
		yield self.stop
		yield self.value

	def __str__(self):
		return '[{start!r}, {stop!r}) -> {value!r}'.format(
			start=self.start,
			stop=self.stop,
			value=self.value,
			)

	def __repr__(self):
		return '{class_name}({start!r}, {stop!r}, {value!r})'.format(
			class_name=self.__class__.__name__,
			start=self.start,
			stop=self.stop,
			value=self.value,
			)


class RangeMapView(Collection):
	"""Base class for views of RangeMaps."""

	__metaclass__ = ABCMeta

	def __init__(self, mapping):
		"""Create a RangeMapView from a RangeMap."""
		self._mapping = mapping

	def __len__(self):
		return len(self._mapping)

	@abstractmethod
	def __iter__(self):
		raise NotImplementedError

	@abstractmethod
	def __contains__(self, item):
		raise NotImplementedError

	def __repr__(self):
		return '{0.__class__.__name__}({0._mapping!r})'.format(self)

	@property
	def mapping(self):
		"""Return the underlying RangeMap."""
		return self._mapping


class RangeMapKeysView(RangeMapView, Set):
	"""A view of the keys that mark the starts of subranges of a RangeMap.

	Since iterating over all the keys is impossible, the view only
	iterates over the keys that start each subrange.
	"""

	def __contains__(self, key):
		return key in self.mapping

	def __iter__(self):
		for mapped_range in self.mapping.ranges():
			yield mapped_range.start


class RangeMapItemsView(RangeMapView, Set):
	"""A view of the items that mark the starts of subranges of a RangeMap.

	Since iterating over all the items is impossible, the view only
	iterates over the items that start each subrange.
	"""

	def __contains__(self, item):
		# TODO should item be a MappedRange instead of a 2-tuple
		key, value = item
		try:
			mapped_value = self.mapping[key]
		except KeyError:
			return False
		else:
			return mapped_value == value

	def __iter__(self):
		for mapped_range in self.mapping.ranges():
			yield (mapped_range.start, mapped_range.value)


class RangeMapValuesView(RangeMapView):
	"""A view on the values that mark the start of subranges of a RangeMap.

	Since iterating over all the values is impossible, the view only
	iterates over the values that start each subrange.
	"""

	def __contains__(self, value):
		for mapped_range in self.mapping.ranges():
			if mapped_range.value == value:
				return True
		return False

	def __iter__(self):
		for mapped_range in self.mapping.ranges():
			yield mapped_range.value


def _check_start_stop(start, stop):
	"""Check that start and stop are valid - orderable and in the right order.

	Raises:
		ValueError: if stop <= start
		TypeError: if unorderable
	"""
	if start is not None and stop is not None and stop <= start:
		raise ValueError('stop must be > start')


def _check_key_slice(key):
	if not isinstance(key, slice):
		raise TypeError('Can only set and delete slices')
	if key.step is not None:
		raise ValueError('Cannot set or delete slices with steps')


class RangeMap(Mapping):
	"""Map ranges of orderable elements to values.

	.. automethod:: __init__
	"""

	def __init__(self, iterable=None, default_value=NOT_SET):
		"""Create a RangeMap.

		A mapping or other iterable can be passed to initialize the RangeMap.
		If mapping is passed, it is interpreted as a mapping from range start
		indices to values.
		If an iterable is passed, each element will define a range in the
		RangeMap and should be formatted (start, stop, value).

		default_value is a an optional keyword argument that will initialize the
		entire RangeMap to that value. Any missing ranges will be mapped to that
		value. However, if ranges are subsequently deleted they will be removed
		and *not* mapped to the default_value.

		Args:
			iterable: A Mapping or an Iterable to initialize from.
			default_value: If passed, the return value for all keys less than the
				least key in mapping or missing ranges in iterable. If no mapping
				or iterable, the return value for all keys.
		"""
		self._keys = [None]
		self._values = [default_value]
		if iterable:
			if isinstance(iterable, Mapping):
				self._init_from_mapping(iterable)
			else:
				self._init_from_iterable(iterable)

	@classmethod
	def from_mapping(cls, mapping):
		"""Create a RangeMap from a mapping of interval starts to values."""
		obj = cls()
		obj._init_from_mapping(mapping)
		return obj

	def _init_from_mapping(self, mapping):
		for key, value in sorted(mapping.items()):
			self.set(value, key)

	@classmethod
	def from_iterable(cls, iterable):
		"""Create a RangeMap from an iterable of tuples defining each range.

		Each element of the iterable is a tuple (start, stop, value).
		"""
		obj = cls()
		obj._init_from_iterable(iterable)
		return obj

	def _init_from_iterable(self, iterable):
		for start, stop, value in iterable:
			self.set(value, start=start, stop=stop)

	def __str__(self):
		range_format = '({range.start}, {range.stop}): {range.value}'
		values = ', '.join([range_format.format(range=r) for r in self.ranges()])
		return 'RangeMap(%s)' % values

	def __repr__(self):
		range_format = '({range.start!r}, {range.stop!r}, {range.value!r})'
		values = ', '.join([range_format.format(range=r) for r in self.ranges()])
		return 'RangeMap([%s])' % values

	def _bisect_left(self, key):
		"""Return the index of the key or the last key < key."""
		if key is None:
			return 0
		else:
			return bisect_left(self._keys, key, lo=1)

	def _bisect_right(self, key):
		"""Return the index of the first key > key."""
		if key is None:
			return 1
		else:
			return bisect_right(self._keys, key, lo=1)

	def ranges(self, start=None, stop=None):
		"""Generate MappedRanges for all mapped ranges.

		Yields:
			MappedRange
		"""
		_check_start_stop(start, stop)
		start_loc = self._bisect_right(start)
		if stop is None:
			stop_loc = len(self._keys)
		else:
			stop_loc = self._bisect_left(stop)
		start_val = self._values[start_loc - 1]
		candidate_keys = [start] + self._keys[start_loc:stop_loc] + [stop]
		candidate_values = [start_val] + self._values[start_loc:stop_loc]
		for i, value in enumerate(candidate_values):
			if value is not NOT_SET:
				start_key = candidate_keys[i]
				stop_key = candidate_keys[i + 1]
				yield MappedRange(start_key, stop_key, value)

	def __contains__(self, key):
		try:
			self._getitem(key)
		except KeyError:
			return False
		else:
			return True

	def __iter__(self):
		for key, value in zip(self._keys, self._values):
			if value is not NOT_SET:
				yield key

	def __bool__(self):
		if len(self._keys) > 1:
			return True
		else:
			return self._values[0] != NOT_SET

	__nonzero__ = __bool__

	def _getitem(self, key):
		"""Get the value for a key (not a slice)."""
		loc = self._bisect_right(key) - 1
		value = self._values[loc]
		if value is NOT_SET:
			raise KeyError(key)
		else:
			return value

	def get(self, key, restval=None):
		"""Get the value of the range containing key, otherwise return restval."""
		try:
			return self._getitem(key)
		except KeyError:
			return restval

	def get_range(self, start=None, stop=None):
		"""Return a RangeMap for the range start to stop.

		Returns:
			A RangeMap
		"""
		return self.from_iterable(self.ranges(start, stop))

	def set(self, value, start=None, stop=None):
		"""Set the range from start to stop to value."""
		_check_start_stop(start, stop)
		# start_index, stop_index will denote the sections we are replacing
		start_index = self._bisect_left(start)
		if start is not None:  # start_index == 0
			prev_value = self._values[start_index - 1]
			if prev_value == value:
				# We're setting a range where the left range has the same
				# value, so create one big range
				start_index -= 1
				start = self._keys[start_index]
		if stop is None:
			new_keys = [start]
			new_values = [value]
			stop_index = len(self._keys)
		else:
			stop_index = self._bisect_right(stop)
			stop_value = self._values[stop_index - 1]
			stop_key = self._keys[stop_index - 1]
			if stop_key == stop and stop_value == value:
				new_keys = [start]
				new_values = [value]
			else:
				new_keys = [start, stop]
				new_values = [value, stop_value]
		self._keys[start_index:stop_index] = new_keys
		self._values[start_index:stop_index] = new_values

	def delete(self, start=None, stop=None):
		"""Delete the range from start to stop from self.

		Raises:
			KeyError: If part of the passed range isn't mapped.
		"""
		_check_start_stop(start, stop)
		start_loc = self._bisect_right(start) - 1
		if stop is None:
			stop_loc = len(self._keys)
		else:
			stop_loc = self._bisect_left(stop)
		for value in self._values[start_loc:stop_loc]:
			if value is NOT_SET:
				raise KeyError((start, stop))
		# this is inefficient, we've already found the sub ranges
		self.set(NOT_SET, start=start, stop=stop)

	def empty(self, start=None, stop=None):
		"""Empty the range from start to stop.

		Like delete, but no Error is raised if the entire range isn't mapped.
		"""
		self.set(NOT_SET, start=start, stop=stop)

	def clear(self):
		"""Remove all elements."""
		self._keys = [None]
		self._values = [NOT_SET]

	@property
	def start(self):
		"""Get the start key of the first range.

		None if RangeMap is empty or unbounded to the left.
		"""
		if self._values[0] is NOT_SET:
			try:
				return self._keys[1]
			except IndexError:
				# This is empty or everything is mapped to a single value
				return None
		else:
			# This is unbounded to the left
			return self._keys[0]

	@property
	def end(self):
		"""Get the stop key of the last range.

		None if RangeMap is empty or unbounded to the right.
		"""
		if self._values[-1] is NOT_SET:
			return self._keys[-1]
		else:
			# This is unbounded to the right
			return None

	def __eq__(self, other):
		if isinstance(other, RangeMap):
			return (
				self._keys == other._keys
				and self._values == other._values
				)
		else:
			return False

	def __getitem__(self, key):
		try:
			_check_key_slice(key)
		except TypeError:
			return self._getitem(key)
		else:
			return self.get_range(key.start, key.stop)

	def __setitem__(self, key, value):
		_check_key_slice(key)
		self.set(value, key.start, key.stop)

	def __delitem__(self, key):
		_check_key_slice(key)
		self.delete(key.start, key.stop)

	def __len__(self):
		count = 0
		for v in self._values:
			if v is not NOT_SET:
				count += 1
		return count

	def keys(self):
		"""Return a view of the keys."""
		return RangeMapKeysView(self)

	def values(self):
		"""Return a view of the values."""
		return RangeMapValuesView(self)

	def items(self):
		"""Return a view of the item pairs."""
		return RangeMapItemsView(self)



"""Sentinel class."""

__all__ = ('Sentinel', 'NOT_SET')


class Sentinel:
	"""A class to create sentinel objects.

	The benefits vs. object() are a good repr it is picklable.

	Inspired by https://pypi.org/project/sentinels/
	"""

	_registry = {}

	def __getnewargs__(self):
		return self._name,

	def __new__(cls, _name):
		"""Find the Sentinel object with name or create a new one."""
		try:
			return cls._registry[_name]
		except KeyError:
			new = super(Sentinel, cls).__new__(cls)
			cls._registry[_name] = new
			return new

	def __init__(self, name):
		super(Sentinel, self).__init__()
		self._name = name

	def __repr__(self):
		return '<%s>' % self._name

	def __bool__(self):
		return False

	def __eq__(self, other):
		if other.__class__ == self.__class__:
			return self._name == other._name
		return False


NOT_SET = Sentinel('not_set')



"""Bag class definitions."""
import heapq
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from collections.abc import Collection, Hashable, Set
from operator import itemgetter

from ._util import deprecated

__all__ = (
	'Bag',
	'BagView',
	'CountsView',
	'UniqueElementsView',
	'bag',
	'frozenbag',
	)


class BagView(Collection):
	"""Base class for bag views."""

	__metaclass__ = ABCMeta
	__slots__ = ('bag', )

	def __init__(self, bag):
		self.bag = bag

	def __repr__(self):
		return '{0.__class__.__name__}({0.bag!r})'.format(self)

	def __len__(self):
		return self.bag.num_unique_elements()

	@abstractmethod
	def __iter__(self):
		raise NotImplementedError

	@abstractmethod
	def __contains__(self, elem):
		raise NotImplementedError


class UniqueElementsView(BagView):
	"""A view for the unique items and their counts in a bag.

	.. versionadded:: 1.0
	"""

	def __iter__(self):
		for elem in self.bag._dict:
			yield elem

	def __contains__(self, elem):
		return elem in self.bag


class CountsView(BagView):
	"""A view for the unique items and their counts in a bag.

	.. versionadded:: 1.0
	"""

	__slots__ = ('bag', )

	def __len__(self):
		return self.bag.num_unique_elements()

	def __iter__(self):
		for elem in self.bag.unique_elements():
			yield elem, self.bag.count(elem)

	def __contains__(self, item):
		elem, count = item
		return self.bag.count(elem) == count


class Bag(Collection):
	"""Base class for bag classes.

	Base class for bag and frozenbag. Is not mutable and not hashable, so there's
	no reason to use this instead of either bag or frozenbag.
	"""

	# Basic object methods

	def __init__(self, iterable=None):
		"""Create a new bag.

		If iterable isn't given, is None or is empty then the bag starts empty.
		Otherwise each element from iterable will be added to the bag
		however many times it appears.

		This runs in O(len(iterable))
		"""
		self._dict = dict()
		self._size = 0
		if iterable:
			if isinstance(iterable, Bag):
				self._dict = iterable._dict.copy()
				self._size = iterable._size
			else:
				for value in iterable:
					self._increment_count(value)

	def _set_count(self, elem, count):
		if count < 0:
			raise ValueError
		self._size += count - self.count(elem)
		if count == 0:
			self._dict.pop(elem, None)
		else:
			self._dict[elem] = count

	def _increment_count(self, elem, count=1):
		self._set_count(elem, self.count(elem) + count)

	@classmethod
	def _from_iterable(cls, it):
		return cls(it)

	def copy(self):
		"""Create a shallow copy of self.

		This runs in O(len(self.num_unique_elements()))
		"""
		out = self._from_iterable(None)
		out._dict = self._dict.copy()
		out._size = self._size
		return out

	def __repr__(self):
		if self._size == 0:
			return '{0}()'.format(self.__class__.__name__)
		else:
			repr_format = '{class_name}({values!r})'
			return repr_format.format(
				class_name=self.__class__.__name__,
				values=tuple(self),
				)

	def __str__(self):
		if self._size == 0:
			return '{class_name}()'.format(class_name=self.__class__.__name__)
		else:
			format_single = '{elem!r}'
			format_mult = '{elem!r}^{mult}'
			strings = []
			for elem, mult in self.counts():
				if mult > 1:
					strings.append(format_mult.format(elem=elem, mult=mult))
				else:
					strings.append(format_single.format(elem=elem))
			return '{%s}' % ', '.join(strings)

	# New public methods (not overriding/implementing anything)

	def num_unique_elements(self):
		"""Return the number of unique elements.

		This runs in O(1) time
		"""
		return len(self._dict)

	def unique_elements(self):
		"""Return a view of unique elements in this bag."""
		return UniqueElementsView(self)

	def count(self, value):
		"""Return the number of value present in this bag.

		If value is not in the bag no Error is raised, instead 0 is returned.

		This runs in O(1) time

		Args:
			value: The element of self to get the count of
		Returns:
			int: The count of value in self
		"""
		return self._dict.get(value, 0)

	@deprecated(
		"Use `heapq.nlargest(n, self.counts(), key=itemgetter(1))` instead or "
		"`sorted(self.counts(), reverse=True, key=itemgetter(1))` for `n=None`",
		'1.0',
		)
	def nlargest(self, n=None):
		"""List the n most common elements and their counts.

		List is from the most
		common to the least.  If n is None, the list all element counts.

		Run time should be O(m log m) where m is len(self)
		Args:
			n (int): The number of elements to return
		"""
		if n is None:
			return sorted(self.counts(), key=itemgetter(1), reverse=True)
		else:
			return heapq.nlargest(n, self.counts(), key=itemgetter(1))

	def counts(self):
		"""Return a view of the unique elements in self and their counts.

		.. versionadded:: 1.1
		"""
		return CountsView(self)

	@classmethod
	def from_mapping(cls, mapping):
		"""Create a bag from a dict of elem->count.

		Each key in the dict is added if the value is > 0.

		Raises:
			ValueError: If any count is < 0.
		"""
		out = cls()
		for elem, count in mapping.items():
			out._set_count(elem, count)
		return out

	# implementing Sized methods

	def __len__(self):
		"""Return the cardinality of the bag.

		This runs in O(1)
		"""
		return self._size

	# implementing Container methods

	def __contains__(self, value):
		"""Return the multiplicity of the element.

		This runs in O(1)
		"""
		return self.count(value)

	# implementing Iterable methods

	def __iter__(self):
		"""Iterate through all elements.

		Multiple copies will be returned if they exist.
		"""
		for value, count in self.counts():
			for _ in range(count):
				yield value

	# Comparison methods

	def issubset(self, other):
		"""Check that every element in self has a count <= in other.

		Args:
			other (Iterable)
		"""
		if not isinstance(other, Bag):
			return self.issubset(frozenbag(other))
		for elem, count in self.counts():
			if not count <= other.count(elem):
				return False
		return True

	def issuperset(self, other):
		"""Check that every element in self has a count >= in other.

		Args:
			other (Iterable)
		"""
		if not isinstance(other, Bag):
			return self.issuperset(bag(other))
		for elem, count in other.counts():
			if not self.count(elem) >= count:
				return False
		return True

	def __le__(self, other):
		if not isinstance(other, Bag):
			return NotImplemented
		return len(self) <= len(other) and self.issubset(other)

	def __lt__(self, other):
		if not isinstance(other, Bag):
			return NotImplemented
		return len(self) < len(other) and self.issubset(other)

	def __gt__(self, other):
		if not isinstance(other, Bag):
			return NotImplemented
		return len(self) > len(other) and self.issuperset(other)

	def __ge__(self, other):
		if not isinstance(other, Bag):
			return NotImplemented
		return len(self) >= len(other) and self.issuperset(other)

	def __eq__(self, other):
		if not isinstance(other, Bag):
			return False
		return self._dict == other._dict

	def __ne__(self, other):
		return not (self == other)

	# Operations - &, |, +, -, ^, * and isdisjoint

	def _iadd(self, other):
		"""Add all of the elements of other to self.

		if isinstance(it, Bag):
				This runs in O(it.num_unique_elements())
		else:
				This runs in O(len(it))
		"""
		if isinstance(other, Bag):
			for elem, count in other.counts():
				self._increment_count(elem, count)
		else:
			for elem in other:
				self._increment_count(elem, 1)
		return self

	def _iand(self, other):
		"""Set multiplicity of each element to the minimum of the two collections.

		if isinstance(other, Bag):
			This runs in O(other.num_unique_elements())
		else:
			This runs in O(len(other))
		"""
		# TODO do we have to create a bag from the other first?
		if not isinstance(other, Bag):
			other = self._from_iterable(other)
		for elem, old_count in set(self.counts()):
			other_count = other.count(elem)
			new_count = min(other_count, old_count)
			self._set_count(elem, new_count)
		return self

	def _ior(self, other):
		"""Set multiplicity of each element to the maximum of the two collections.

		if isinstance(other, Bag):
			This runs in O(other.num_unique_elements())
		else:
			This runs in O(len(other))
		"""
		# TODO do we have to create a bag from the other first?
		if not isinstance(other, Bag):
			other = self._from_iterable(other)
		for elem, other_count in other.counts():
			old_count = self.count(elem)
			new_count = max(other_count, old_count)
			self._set_count(elem, new_count)
		return self

	def _ixor(self, other):
		"""Set self to the symmetric difference between the sets.

		if isinstance(other, Bag):
			This runs in O(other.num_unique_elements())
		else:
			This runs in O(len(other))
		"""
		if isinstance(other, Bag):
			for elem, other_count in other.counts():
				count = abs(self.count(elem) - other_count)
				self._set_count(elem, count)
		else:
			# Let a = self.count(elem) and b = other.count(elem)
			# if a >= b then elem is removed from self b times leaving a - b
			# if a < b then elem is removed from self a times then added (b - a)
			# times leaving a - a + (b - a) = b - a
			for elem in other:
				try:
					self._increment_count(elem, -1)
				except ValueError:
					self._increment_count(elem, 1)
		return self

	def _isub(self, other):
		"""Discard the elements of other from self.

		if isinstance(it, Bag):
			This runs in O(it.num_unique_elements())
		else:
			This runs in O(len(it))
		"""
		if isinstance(other, Bag):
			for elem, other_count in other.counts():
				try:
					self._increment_count(elem, -other_count)
				except ValueError:
					self._set_count(elem, 0)
		else:
			for elem in other:
				try:
					self._increment_count(elem, -1)
				except ValueError:
					pass
		return self

	def __and__(self, other):
		"""Intersection is the minimum of corresponding counts.

		This runs in O(l + n) where:
			* n is self.num_unique_elements()
			* `l = 1` if other is a bag else `l = len(other)`

		"""
		return self.copy()._iand(other)

	def isdisjoint(self, other):
		"""Return if this bag is disjoint with the passed collection.

		This runs in O(len(other))
		"""
		for value in other:
			if value in self:
				return False
		return True

	def __or__(self, other):
		"""Union is the maximum of all elements.

		This runs in O(m + n) where:
			* `n = self.num_unique_elements()`
			* m = other.num_unique_elements() if other is a bag else m = len(other)
		"""
		return self.copy()._ior(other)

	def __add__(self, other):
		"""Return a new bag also containing all the elements of other.

		self + other = self & other + self | other

		This runs in O(m + n) where:
			* n is self.num_unique_elements()
			* m is len(other)

		Args:
			other (Iterable): elements to add to self

		"""
		return self.copy()._iadd(other)

	def __sub__(self, other):
		"""Difference between the sets.

		For normal sets this is all x s.t. x in self and x not in other.
		For bags this is count(x) = max(0, self.count(x)-other.count(x))

		This runs in O(m + n) where:
			* n is self.num_unique_elements()
			* m is len(other)

		Args:
			other (Iterable): elements to remove

		"""
		return self.copy()._isub(other)

	def __mul__(self, other):
		"""Cartesian product with other."""
		return self.product(other)

	def product(self, other, operator=None):
		"""Cartesian product of the two sets.

		Optionally, pass an operator to combine elements instead of creating a
		tuple.

		This should run in O(m*n+l) where:
			* `m` is the number of unique elements in `self`
			* `n` is the number of unique elements in `other`
			* `l` is 0 if `other` is a bag, else `l` is the `len(other)`

		Args:
			other (Iterable): The iterable to take the product with.
			operator (Callable): A function that accepts an element from self
				and other and returns a combined value to include in the return
				value.
		"""
		if not isinstance(other, Bag):
			other = self._from_iterable(other)
		values = defaultdict(int)
		for elem, count in self.counts():
			for other_elem, other_count in other.counts():
				if operator:
					new_elem = operator(elem, other_elem)
				else:
					new_elem = (elem, other_elem)
				new_count = count * other_count
				values[new_elem] += new_count
		return self.from_mapping(values)

	def __xor__(self, other):
		"""Symmetric difference between the sets.

		other can be any iterable.

		This runs in O(m + n) where:
			m = len(self)
			n = len(other)
		"""
		return self.copy()._ixor(other)


class bag(Bag):
	"""bag is a mutable unhashable bag.

	.. automethod:: __init__
	"""

	def pop(self):
		"""Remove and return an element of self."""
		# TODO can this be done more efficiently (no need to create an iterator)?
		it = iter(self)
		try:
			value = next(it)
		except StopIteration:
			raise KeyError('pop from an empty bag')
		self.remove(value)
		return value

	def add(self, elem):
		"""Add elem to self."""
		self._increment_count(elem)

	def discard(self, elem):
		"""Remove elem from this bag, silent if it isn't present."""
		try:
			self.remove(elem)
		except ValueError:
			pass

	def remove(self, elem):
		"""Remove elem from this bag, raising a ValueError if it isn't present.

		Args:
			elem: object to remove from self
		Raises:
			ValueError: if the elem isn't present
		"""
		self._increment_count(elem, -1)

	def discard_all(self, other):
		"""Discard all of the elems from other."""
		self._isub(other)

	def remove_all(self, other):
		"""Remove all of the elems from other.

		Raises a ValueError if the multiplicity of any elem in other is greater
		than in self.
		"""
		if not self.issuperset(other):
			raise ValueError('Passed collection is not a subset of this bag')
		self.discard_all(other)

	def clear(self):
		"""Remove all elements from this bag."""
		self._dict = dict()
		self._size = 0

	# In-place operations

	__ior__ = Bag._ior
	__iand__ = Bag._iand
	__ixor__ = Bag._ixor
	__isub__ = Bag._isub
	__iadd__ = Bag._iadd


class frozenbag(Bag, Hashable):
	"""frozenbag is an immutable, hashable bag.

	.. automethod:: __init__
	"""

	def __hash__(self):
		"""Compute the hash value of a frozenbag."""
		if not hasattr(self, '_hash_value'):
			self._hash_value = Set._hash(self)
		return self._hash_value



"""IndexedDict class definition.

.. versionadded:: 1.1
"""
from collections.abc import MutableMapping

from ._util import deprecation_warning
from .sentinel import NOT_SET

__all__ = ('IndexedDict', )

# TODO these should be ValueErrors
KEY_AND_INDEX_ERROR = TypeError(
	"Specifying both `key` and `index` is not allowed")
KEY_EQ_INDEX_ERROR = TypeError(
	"Exactly one of `key` and `index` must be specified")


class IndexedDict(MutableMapping):
	"""A Mapping that preserves insertion order and allows access by item index.

	The API is an extension of OrderedDict.

	.. automethod:: __init__
	"""

	def __init__(self, iterable=None, **kwargs):
		"""Create an IndexedDict and initialize it like a dict."""
		self._dict = {}  # key -> (index, value)
		self._list = []  # index -> (key, value)
		self.update(iterable or [], **kwargs)

	def clear(self):
		"""Remove all items."""
		self._dict = {}
		self._list = []

	def get(self, key=NOT_SET, index=NOT_SET, default=NOT_SET, d=NOT_SET):
		"""Return value with given `key` or `index`.

		If no value is found, return `default` (`None` by default).

		.. deprecated :: 1.1

		The `d` parameter has been renamed `default`. `d` will be removed in
		some future version.

		Args:
			key: The key of the value to get
			index: The index of the value to get
			default: The value to return if `key` is not found or `index` is
				out of bounds. If it is NOT_SET, None is returned.
			d: DEPRECATED: Old parameter name for `default`

		"""
		if d is not NOT_SET:
			if default is not NOT_SET:
				raise ValueError('Specified default and d')
			deprecation_warning(
				"IndexedDict.pop parameter 'd' has been renamed to 'default'"
				)
			default = d
		if default is NOT_SET:
			default = None

		if index is NOT_SET and key is not NOT_SET:
			try:
				index, value = self._dict[key]
			except KeyError:
				return default
			else:
				return value
		elif index is not NOT_SET and key is NOT_SET:
			try:
				key, value = self._list[index]
			except IndexError:
				return default
			else:
				return value
		else:
			raise KEY_EQ_INDEX_ERROR

	def pop(self, key=NOT_SET, index=NOT_SET, default=NOT_SET, d=NOT_SET):
		"""Remove and return value.

		Optionally, specify the `key` or `index` of the value to pop.
		If `key` is specified and is not found a `KeyError` is raised unless
		`default` is specified. Likewise, if `index` is specified that is out of
		bounds, an `IndexError` is raised unless `default` is specified.

		Both `index` and `key` cannot be specified. If neither is specified,
		then the last value is popped.

		This is generally O(N) unless removing last item, then O(1).

		.. deprecated :: 1.1

		The `d` parameter has been renamed `default`. `d` will be removed in
		some future version.

		Args:
			key: The key of the value to pop
			index: The index of the value to pop
			default: The value to return if the key is not found or the index is
				out of bounds
			d: DEPRECATED: Old parameter name for `default`

		"""
		if d is not NOT_SET:
			if default is not NOT_SET:
				raise ValueError('Specified default and d')
			deprecation_warning(
				"IndexedDict.pop parameter 'd' has been renamed to 'default'"
				)
			default = d

		has_default = default is not NOT_SET
		if index is NOT_SET and key is not NOT_SET:
			index, value = self._pop_key(key, has_default)
		elif key is NOT_SET:
			key, index, value = self._pop_index(index, has_default)
		else:
			raise KEY_AND_INDEX_ERROR

		if index is None:
			return default
		else:
			self._fix_indices_after_delete(index)
			return value

	def _pop_key(self, key, has_default):
		"""Remove an element by key."""
		try:
			index, value = self._dict.pop(key)
		except KeyError:
			if has_default:
				return None, None
			else:
				raise
		key2, value2 = self._list.pop(index)
		assert key is key2
		assert value is value2

		return index, value

	def _pop_index(self, index, has_default):
		"""Remove an element by index, or last element."""
		try:
			if index is NOT_SET:
				index = len(self._list) - 1
				key, value = self._list.pop()
			else:
				key, value = self._list.pop(index)
				if index < 0:
					index += len(self._list) + 1
		except IndexError:
			if has_default:
				return None, None, None
			else:
				raise
		index2, value2 = self._dict.pop(key)
		assert index == index2
		assert value is value2

		return key, index, value

	def fast_pop(self, key=NOT_SET, index=NOT_SET):
		"""Pop a specific item quickly by swapping it to the end.

		Remove value with given key or index (last item by default) fast
		by swapping it to the last place first.

		Changes order of the remaining items (item that used to be last goes to
		the popped location).
		Returns tuple of (poped_value, new_moved_index, moved_key, moved_value).
		If key is not found raises KeyError or IndexError.

		Runs in O(1).
		"""
		if index is NOT_SET and key is not NOT_SET:
			index, popped_value = self._dict.pop(key)
		elif key is NOT_SET:
			if index is NOT_SET:
				index = len(self._list) - 1
				key, popped_value2 = self._list[-1]
			else:
				key, popped_value2 = self._list[index]
				if index < 0:
					index += len(self._list)
			index2, popped_value = self._dict.pop(key)
			assert index == index2
		else:
			raise KEY_AND_INDEX_ERROR

		if key == self._list[-1][0]:
			# The item we're removing happens to be the last in the list,
			# no swapping needed
			_, popped_value2 = self._list.pop()
			assert popped_value is popped_value2
			return popped_value, len(self._list), key, popped_value
		else:
			# Swap the last item onto the deleted spot and
			# pop the last item from the list
			self._list[index] = self._list[-1]
			moved_key, moved_value = self._list.pop()
			self._dict[moved_key] = (index, moved_value)
			return popped_value, index, moved_key, moved_value

	def popitem(self, last=NOT_SET, *, key=NOT_SET, index=NOT_SET):
		"""Remove and return a (key, value) tuple.

		By default, the last item is popped.
		Optionally, specify the `key` or `index` of the value to pop.
		The `last` parameter is included to match the OrderedDict API. If `last`
		is passed then the first or last item is returned based on its
		truthiness.

		At most one of `index`, `last` and `key` can be specified.

		This is generally O(N) unless removing last item, then O(1).

		Args:
			key: The key of the value to pop
			index: The index of the value to pop
			last: Whether or not to pip the last item

		Raises:
			KeyError: If the dictionary is empty or a key is specified that is
				not present
			IndexError: If `index` is specified and is out of bounds
			ValueError: If more than one of `last`, `index` and `key` are
				specified

		"""
		if not self:
			raise KeyError('IndexedDict is empty')
		if sum(x is not NOT_SET for x in (last, key, index)) > 1:
			raise ValueError(
				"Cannot specify more than one of key, index and last"
				)
		if key is not NOT_SET:
			index, value = self._pop_key(key=key, has_default=False)
		else:
			if last is not NOT_SET:
				index = -1 if last else 0
			if index is NOT_SET:
				index = -1
			key, index, value = self._pop_index(index, has_default=False)
		self._fix_indices_after_delete(starting_index=index)
		return key, value

	def move_to_end(self, key=NOT_SET, index=NOT_SET, last=True):
		"""Move an existing element to the end (or beginning if last==False).

		Runs in O(N).
		"""
		if index is NOT_SET and key is not NOT_SET:
			index, value = self._dict[key]
		elif index is not NOT_SET and key is NOT_SET:
			key, value = self._list[index]

			# Normalize index
			if index < 0:
				index += len(self._list)
		else:
			raise KEY_EQ_INDEX_ERROR

		if last:
			index_range = range(len(self._list) - 1, index - 1, -1)
			self._dict[key] = (len(self._list) - 1, value)
		else:
			index_range = range(index + 1)
			self._dict[key] = (0, value)

		previous = (key, value)
		for i in index_range:
			self._dict[previous[0]] = i, previous[1]
			previous, self._list[i] = self._list[i], previous

	def copy(self):
		"""Return a shallow copy."""
		ret = IndexedDict()
		ret._dict = self._dict.copy()
		ret._list = list(self._list)
		return ret

	def index(self, key):
		"""Return index of a record with given key.

		Runs in O(1).
		"""
		return self._dict[key][0]

	def key(self, index):
		"""Return key of a record at given index.

		Runs in O(1).
		"""
		return self._list[index][0]

	def __len__(self):
		"""Return number of elements stored."""
		return len(self._list)

	def __repr__(self):
		return "{class_name}({data})".format(
			class_name=self.__class__.__name__,
			data=repr(self._list),
			)

	def __str__(self):
		# When Python 3.5 support is dropped, we can rely on dict order and this
		# can be simplified to:
		# return "{class_name}({data})".format(
		# 	class_name=self.__class__.__name__,
		# 	data=repr(dict(self)),
		# 	)
		data = ', '.join(
			'{k!r}: {v!r}'.format(k=k, v=v)
			for k, v in self.items()
			)
		return "{class_name}({{{data}}})".format(
			class_name=self.__class__.__name__,
			data=data,
			)

	def __getitem__(self, key):
		"""Return value corresponding to given key.

		Raises KeyError when the key is not present in the mapping.

		Runs in O(1).
		"""
		return self._dict[key][1]

	def __setitem__(self, key, value):
		"""Set item with given key to given value.

		If the key is already present in the mapping its order is unchanged,
		if it is not then it's added to the last place.

		Runs in O(1).
		"""
		if key in self._dict:
			index, old_value1 = self._dict[key]
			self._list[index] = key, value
		else:
			index = len(self._list)
			self._list.append((key, value))
		self._dict[key] = index, value

	def __delitem__(self, key):
		"""Remove item with given key from the mapping.

		Runs in O(n), unless removing last item, then in O(1).
		"""
		index, value = self._dict.pop(key)
		key2, value2 = self._list.pop(index)
		assert key == key2
		assert value is value2

		self._fix_indices_after_delete(index)

	def __contains__(self, key):
		"""Check if a key is present in the mapping.

		Runs in O(1).
		"""
		return key in self._dict

	def __iter__(self):
		"""Return iterator over the keys of the mapping in order."""
		return (item[0] for item in self._list)

	def _fix_indices_after_delete(self, starting_index=0):
		for i, (k, v) in enumerate(self._list[starting_index:], starting_index):
			self._dict[k] = (i, v)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['collections_extended']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'collections-extended',
    'version': '2.0.0',
    'description': 'Extra Python Collections - bags (multisets) and setlists (ordered sets)',
    'long_description': 'README\n######\n\n.. image:: https://coveralls.io/repos/github/mlenzen/collections-extended/badge.svg?branch=master\n\t:target: https://coveralls.io/github/mlenzen/collections-extended?branch=master\n\t:alt: Coverage\n\n\n.. image:: https://pepy.tech/badge/collections-extended/week\n\t:target: https://pepy.tech/project/collections-extended/\n\t:alt: Downloads\n\n\nDocumentation:\n\thttp://collections-extended.lenzm.net/\nGitHub:\n\thttps://github.com/mlenzen/collections-extended\nPyPI:\n\thttps://pypi.python.org/pypi/collections-extended\n\nOverview\n========\n\n``collections_extended`` is a Python module providing\na ``bag`` class, AKA **multiset**,\na ``setlist`` class, which is a **unique list** or **ordered set**,\na ``bijection`` class, ``RangeMap`` which is a mapping from ranges to values and\na ``IndexedDict`` class, which is an ordered mapping whose elements can be accessed using index,\nin addition to key.\nThere are also frozen (hashable) varieties of bags and setlists.\n\nCompatible with and tested against Python 3.6, 3.7, 3.8, 3.9 & PyPy3.\n\nGetting Started\n===============\n\n.. code-block:: python\n\n\t>>> from collections_extended import bag, setlist, bijection, RangeMap, IndexedDict\n\t>>> from datetime import date\n\t>>> b = bag(\'abracadabra\')\n\t>>> b.count(\'a\')\n\t5\n\t>>> b.remove(\'a\')\n\t>>> b.count(\'a\')\n\t4\n\t>>> \'a\' in b\n\tTrue\n\t>>> b.count(\'d\')\n\t1\n\t>>> b.remove(\'d\')\n\t>>> b.count(\'d\')\n\t0\n\t>>> \'d\' in b\n\tFalse\n\n\t>>> sl = setlist(\'abracadabra\')\n\t>>> sl\n\tsetlist((\'a\', \'b\', \'r\', \'c\', \'d\'))\n\t>>> sl[3]\n\t\'c\'\n\t>>> sl[-1]\n\t\'d\'\n\t>>> \'r\' in sl  # testing for inclusion is fast\n\tTrue\n\t>>> sl.index(\'d\')  # so is finding the index of an element\n\t4\n\t>>> sl.insert(1, \'d\')  # inserting an element already in raises a ValueError\n\tTraceback (most recent call last):\n\t...\n\t\traise ValueError\n\tValueError\n\t>>> sl.index(\'d\')\n\t4\n\n\t>>> bij = bijection({\'a\': 1, \'b\': 2, \'c\': 3})\n\t>>> bij.inverse[2]\n\t\'b\'\n\t>>> bij[\'a\'] = 2\n\t>>> bij == bijection({\'a\': 2, \'c\': 3})\n\tTrue\n\t>>> bij.inverse[1] = \'a\'\n\t>>> bij == bijection({\'a\': 1, \'c\': 3})\n\tTrue\n\n\t>>> version = RangeMap()\n\t>>> version[date(2017, 10, 20): date(2017, 10, 27)] = \'0.10.1\'\n\t>>> version[date(2017, 10, 27): date(2018, 2, 14)] = \'1.0.0\'\n\t>>> version[date(2018, 2, 14):] = \'1.0.1\'\n\t>>> version[date(2017, 10, 24)]\n\t\'0.10.1\'\n\t>>> version[date(2018, 7, 1)]\n\t\'1.0.1\'\n\t>>> version[date(2018, 6, 30):] = \'1.0.2\'\n\t>>> version[date(2018, 7, 1)]\n\t\'1.0.2\'\n\n\t>>> idict = IndexedDict()\n\t>>> idict[\'a\'] = "A"\n\t>>> idict[\'b\'] = "B"\n\t>>> idict[\'c\'] = "C"\n\t>>> idict.get(key=\'a\')\n\t\'A\'\n\t>>> idict.get(index=2)\n\t\'C\'\n\t>>> idict.index(\'b\')\n\t1\n\nInstallation\n============\n\n``pip install collections-extended``\n\nUsage\n=====\n\t``from collections_extended import bag, frozenbag, setlist, frozensetlist, bijection``\n\nClasses\n=======\nThere are seven new classes provided:\n\nBags\n----\nbag\n\tThis is a bag AKA multiset.\nfrozenbag\n\tThis is a frozen (hashable) version of a bag.\n\nSetlists\n--------\nsetlist\n\tAn ordered set or a list of unique elements depending on how you look at it.\nfrozensetlist\n\tThis is a frozen (hashable) version of a setlist.\n\nMappings\n--------\nbijection\n\tA one-to-one mapping.\nRangeMap\n\tA mapping from ranges (of numbers/dates/etc)\nIndexedDict\n\tA mapping that keeps insertion order and allows access by index.\n\nPython 2\n--------\n\nThe package no longer supports Python 2. The last version to support\nPython 2.7, 3.4 & 3.5 was 1.0. No new feature releases will be done for 1.x but\nany significant bugs that come up may be fixed.\n\n:Author: Michael Lenzen\n:Copyright: 2021 Michael Lenzen\n:License: Apache License, Version 2.0\n:Project Homepage: https://github.com/mlenzen/collections-extended\n',
    'author': 'Michael Lenzen',
    'author_email': 'm.lenzen@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mlenzen/collections-extended',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
